% check that our TE actually works on some synthetic series

%% prepare the data
len = 300;

x = 1:len;

dy = randn(1, len) * 0.1;
y = zeros(1, len);
y(1) = 0;
for i = 2:len
	y(i) = y(i-1)*0.9 + dy(i);
end

yd = [randn(1, 5)*0.1, y(1:end-5)] + randn(1,len)*0.01; % shifted by 5 with small noise

yr = yd(randperm(len));


%% print it
figure(123)
hold off
plot(x, y, 'k');
hold on
plot(x, yd, 'b');
plot(x, yr, 'r', 'LineWidth', 0.1);
lh = legend('$X$', '$Y$', '$\bar Y$');
set(lh, 'Interpreter', 'LaTeX');
drawnow

%% compute the stuff

n = 4; % length of the window
dt = 2; % offset to the past

X_t1 = [];
X_t = [];
Yd_t1 = [];
Yd_t = [];
Yr_t1 = [];
Yr_t = [];

sigmasd = [];
sigmasr = [];
sigmasdb = [];
sigmasrb = [];

for t = n+1+dt : len
	
	x_t1 = y(t);
	x_t = y(t-n-dt:t-1-dt);
	yd_t1 = yd(t);
	yd_t = yd(t-n:t-1);
	yr_t1 = yr(t);
	yr_t = yr(t-n:t-1);
	
	X_t1 = [X_t1 x_t1'];
	X_t = [X_t x_t'];
	Yd_t1 = [Yd_t1 yd_t1'];
	Yd_t = [Yd_t yd_t'];
	Yr_t1 = [Yr_t1 yr_t1'];
	Yr_t = [Yr_t yr_t'];
	
	if mod(t, 100) == 1 % regular reset to escape possible local minimum
		sigmasd = [];
		sigmasr = [];
		sigmasdb = [];
		sigmasrb = [];
	end
	
	if length(Yd_t1) >= 50 + (2*n+1)
		% predicting X->Y (y->yd)
		[TEd(t), sigmasd] = estimTransferEn(Yd_t1, Yd_t, X_t, sigmasd);
		% predicting X->Yr (y->yr)
		[TEr(t), sigmasr] = estimTransferEn(Yr_t1, Yr_t, X_t, sigmasr);
		
		% this way reversed, there will be some noticeable information trans
		% predicting Y->X (yd->y)
		[TEdb(t), sigmasdb] = estimTransferEn(X_t1, X_t, Yd_t, sigmasdb);
		% predicting Yr->X (yr->y)
		[TErb(t), sigmasrb] = estimTransferEn(X_t1, X_t, Yr_t, sigmasrb);
		fprintf('Time t = %d done... (%.1f/%.1f/%.1f/%.1f)\n', t, TEd(t), TEr(t), TEdb(t), TErb(t));
	end
	drawnow
end


%% plot the results

figure(124)
hold off
plot(TEd, 'b', 'LineWidth', 2);
hold on
plot(TEr, 'r');
plot(TEdb, 'g');
plot(TErb, 'm');

title('TE');
lh = legend('$X \rightarrow Y$', '$X \rightarrow \bar Y$', '$Y \rightarrow X$', '$\bar Y \rightarrow X$',...
		'Location', 'E');
set(lh, 'Interpreter', 'LaTeX');
xlabel('frame #');
ylabel('entropy (b)');



%% statistical test

start = find(TEd == 0, 1, 'last') + 1;

P = ones(1,len);
Pb = ones(1,len);

for i = start+1:len
	start2 = max(start, i-99);
	P(i) = ttestWelch(TEd(start2:i), TEr(start2:i), true);
	Pb(i) = ttestWelch(TEdb(start2:i), TErb(start2:i), true);
end

%% plot the permutation test results

figure(125);
hold off
semilogy(P, 'b', 'LineWidth', 2);
hold on
semilogy(Pb, 'r')

lh = legend('$X \rightarrow Y$ vs. $X \rightarrow \bar Y$', '$Y \rightarrow X$ vs. $\bar Y \rightarrow X$',...
		'Location', 'E');
set(lh, 'Interpreter', 'LaTeX');

xlabel('frame #');
ylabel('p-value (-)');



%% find where it crosses a threshold

pvalThr = 1e-4;

for i = 1:len
	if P(i) < pvalThr
		break
	end
end

if i == len
	selectedFrame = 0;
	fprintf('No significant TE was found.\n');
else
	selectedFrame = i;
	fprintf('The first significant TE was found at frame %i.\n', selectedFrame);
end


%% do the grid search

dtm = 0;
dtM = 7;
nm = 1;
nM = 5;

TEgrid = zeros(dtM-dtm+1, nM-nm+1);

for dti = dtm:dtM
	for ni = nm:nM
		[X_t, Y_t, Y_t1] = getPoints(y, yd, ni, dti, selectedFrame);
		TEgrid(dti-dtm+1, ni-nm+1) = estimTransferEn(Y_t1, Y_t, X_t);
		fprintf('dt = %d, n = %d done...\n', dti, ni);
	end
	
end

relImpr = relativeImprovement(TEgrid);




%% plot the grid of TE

riThr = 0.1;


figure(126);
hold off
b3h = bar3(TEgrid, 0.6);
hold on
[dti, ni] = find(relImpr > riThr);
plot3(ni, dti, 0.1 + TEgrid(sub2ind(size(TEgrid), dti, ni)), 'r*')
lh = xlabel('$n$');
set(lh, 'Interpreter', 'LaTeX');
lh = ylabel('$\Delta t$');
set(lh, 'Interpreter', 'LaTeX');
zlabel('TE (b)');
colorbar
colormap jet
view([-60, 40]);
xlim([0.5, size(TEgrid, 2) + 0.5]);
ylim([0.5, size(TEgrid, 1) + 0.5]);
set(gca, 'CLim', [-1 1]);
axis equal
set(gca, 'XTick', 1:nM-nm+1, 'XTickLabel', nm:nM);
set(gca, 'YTick', 1:dtM-dtm+1, 'YTickLabel', dtm:dtM);


for i = 1:length(b3h)
	set(b3h(i), 'CData', kron(relImpr(:,i), ones(6,4)) );
end


%% get the optimal parameters

TEf = TEgrid(relImpr > riThr);
[~, ind] = max(TEf);
[dti, ni] = find(relImpr > riThr);

dtOptimal = dti(ind) + dtm - 1;
nOptimal = ni(ind) + nm - 1;

fprintf('Optimal parameters are: dt = %d, n = %i.\n', dtOptimal, nOptimal);



