'''
Script to train a gaussian process for regressing points on an objects surface.

command line args:
-i input filepath
-b bias level
-hyper train hyperparameters (0 is just recondition, 1 is full retrain, 2 is retrain only noise level)
-d num_dimensions
-n num_pts
-u flag to input point uncertainty
-v verbose
-f filename to save out GP to

return codes:
0    success
1    invalid csv file dimensionality
2    File IO failure

Created on Jun 11, 2014

@author: Simon
'''

import argparse
from numpy import *
import GPy
import sys
import random
import csv

verbose_level=0

def vprint(text,level=1):
    """possibly print out a message, depending on the requested verbosity level"""
    if level<=verbose_level:
        for item in text:
            print "verbose: ",item
        print "\n"


def load_data(pathI,pathO):
    #read csv file, should be n by d+1
    vprint(["Loading csv files from ", pathI, pathO],1)
    try:
        csvfileI=open(pathI)
        csvfileO=open(pathO)
    except:
        raise Exception("Failed to open CSV files from " + pathI+" and "+pathO)

    #create a list of list of words and creat a numpy array
    image_seq_reader=csv.reader(csvfileI,delimiter=',')
    object_seq_reader=csv.reader(csvfileO,delimiter=',')
    image_seq=[]
    object_seq=[]
    for row in image_seq_reader:
        image_seq.append([])
        for item in row:
            image_seq[-1].append(float(item))

    for row in object_seq_reader:
        object_seq.append([])
        for item in row:
            object_seq[-1].append(float(item))
    
    vprint(["Loaded image_seq series had shape",shape(image_seq)],1)
    vprint(["Loaded object_seq series had shape",shape(object_seq)],1)
    vprint(["image_seq range from ",min(min(image_seq)),"to",max(max(image_seq))],1)
    vprint(["object_seq range from ",min(min(object_seq)),"to",max(max(object_seq))],1)

    #frame numbers for each set of results
    frame_numbers=array(range(size(image_seq,1)))
    frame_numbers=frame_numbers[:,None]
    
    return [frame_numbers, array(image_seq).transpose(), array(object_seq).transpose()]



def create_GP(image_frame_numbers, object_frame_numbers, image_seq, object_seq, lagStd):
    dims=size(image_seq,1) #input dims
    
    #reshape variables    
    index = vstack((zeros_like(image_seq), ones_like(object_seq)))
    stacked_frame_nums = hstack((vstack((image_frame_numbers, object_frame_numbers)), index))

    #build a suitable set of observed variables
    sequences = vstack((image_seq,object_seq))
    
    W11=array([[1],[0]])
    W22=array([[0],[1]])
    #rbf for X vs X and Y vs Y
    k11 = ( GPy.kern.rbf(dims) + GPy.kern.bias(dims) ).prod(
                    GPy.kern.coregionalize(2*dims,W=W11,kappa=array([0,0])),
                    tensor=True)
    k22 = ( GPy.kern.rbf(dims) + GPy.kern.bias(dims) ).prod(
                    GPy.kern.coregionalize(2*dims,W=W22,kappa=array([0,0])),
                    tensor=True)
    
    #offset RBF for X vs Y 
    k12 = ( GPy.kern.rbf(dims) + GPy.kern.bias(dims) ).prod(
                    GPy.kern.coregionalize(2*dims,2*dims,kappa=array([0,0])),
                    tensor=True)
    #print k12
    k12[".*coregion_W0_0"]=0.7
    k12[".*coregion_W1_0"]=0.7
    k12[".*coregion_W0_1"]=0.7
    k12[".*coregion_W1_1"]=0.7
    
    k_joint = k11 + k22 + k12
    
    #create models    
    model = GPy.models.GPRegression(stacked_frame_nums,sequences,kernel=k_joint)
    
    model[".*lengthscale"]=lagStd
    model.ensure_default_constraints()
    
    model.constrain_fixed(".*coregion_kappa")
    #fix the on diagonals to 1
    model.constrain_fixed(".*coregion_3_coregion_W0_0")
    model.constrain_fixed(".*coregion_3_coregion_W1_0")
    
    model.constrain_fixed(".*coregion_2_coregion_W0_0")
    model.constrain_fixed(".*coregion_2_coregion_W1_0")

    model.constrain_fixed(".*coregion_1_coregion_W0_0")
    model.constrain_fixed(".*coregion_1_coregion_W1_1")
    
    return model


def train_model(image_frame_numbers, object_frame_numbers, image_seq, object_seq, lagStd):
    model = create_GP(image_frame_numbers, object_frame_numbers, image_seq, object_seq, lagStd)
    vprint(['beginning optimization of gp',model])
    model.optimize_restarts(max_iters=100,ftol=1e-2,messages=verbose_level>0,robust=True,parallel=True,num_restarts=5)
    return model



def test_model(model, frame_numbers):
    """Take a gaussian process model and some numpy arrays of test data, and calculate the regression error"""
    
    #get predictions
    vprint(['beginning prediction on test data',model])
    index = ones_like(frame_numbers)
    stacked_frame_numbers = hstack((frame_numbers,index))
    mean,var,u,l=model.predict(stacked_frame_numbers)

    return [mean, var]


if __name__ == '__main__':

    random.seed(100) #if numpy isn't seeded at all, the output is always the same anyway, but lets explicitly seed it with a constant

    #read command line args
    parser = argparse.ArgumentParser(description='Script to train a gaussian process for regressing points on an objects surface.')
    parser.add_argument('-fi',nargs=1,type=str,required=True,help="path to the input CSV file (image data)")
    parser.add_argument('-fo',nargs=1,type=str,required=True,help="path to the input CSV file (object data)")
    parser.add_argument('-o',nargs=1,type=str,required=True,help="path to the output CSV file (object data)")
    parser.add_argument('-n',nargs=1,type=int,default=[0],help="number of input points")
    parser.add_argument('-d',nargs=1,type=int,default=[1],help="dimensionality of input points (default=1)")
    parser.add_argument('-v',nargs=1,type=int,default=[0],help="print out verbose information (higher level prints more, default=0)")
    parser.add_argument('-dt',nargs=1,type=float,default=[0],help="lag of image sequence behind object sequence")
    parser.add_argument('-win',nargs=1,type=float,default=[2],help="length of influencing window of the image signal")
    parser.add_argument('-frames',nargs=2,type=int,default=[2, 0],help="range of frames to process, inclusive, 1-based")
    args=parser.parse_args()

    verbose_level=args.v[0]
    if verbose_level > 0 :
        import time
        start=time.time()
    vprint(["Command line arguments were",args])

    #read csv files, should be n by d
    [frame_numbers, image_seq, object_seq]=load_data(args.fi[0], args.fo[0])
    if args.n[0] != 0 and size(image_seq,0) != args.n[0]:
        print "ERROR: Command line arguments specified ",args.n[0],"image points, but read",size(image_seq,0),"from file at",args.fi
        sys.exit(1)
    if args.n[0] != 0 and size(object_seq,0) != args.n[0]:
        print "ERROR: Command line arguments specified ",args.n[0],"object points, but read",size(object_seq,0),"from file at",args.fo
        sys.exit(1)
    if size(image_seq,1) != args.d[0]:
        print "ERROR: Command line arguments specified ",args.d[0],"image dimensions, but read",size(image_seq,1),"from file at",args.fi
        sys.exit(1)
    if size(object_seq,1) != args.d[0]:
        print "ERROR: Command line arguments specified ",args.d[0],"object dimensions, but read",size(object_seq,1),"from file at",args.fo
        sys.exit(1)
    
    preds = array([])
    uncs = array([])
    lag = args.dt[0]
    lagStd = (args.win[0]/2)
    lag += lagStd # centre of the window
    
    frameRange = args.frames
    frameRange[0] -= 1
    if frameRange[1] == 0:
        frameRange[1] = size(frame_numbers,0)
    
    for frame_num in range(frameRange[0], frameRange[1]):
            print "\n\nstarting frame",frame_num,"of",size(frame_numbers,0)-1,"\n\n"
            #train data
            firstFrameI = max(0, frame_num + max(0,-lag) - 100)
            firstFrameO = max(0, frame_num - 100)
            image_train_frames=frame_numbers[firstFrameI:min(frame_num + 1 + max(0,-lag), size(frame_numbers,0)),:]+lag;
            image_train_seqs=image_seq[firstFrameI:min(frame_num + 1 + max(0,-lag), size(frame_numbers,0)),:];
            object_train_frames=frame_numbers[firstFrameO:frame_num,:];
            object_train_seqs=object_seq[firstFrameO:frame_num,:];
    
            vprint(["Training model with ",size(image_train_frames,0),"samples and ",size(object_train_frames,0)],1)
            model=train_model(image_train_frames, object_train_frames, image_train_seqs, object_train_seqs, lagStd)
            vprint(["Newly trained model had values ",model])
            
            #test model
            test_frames=frame_numbers[frame_num,:];
            [pred, unc] = test_model(model, test_frames)
            preds = append(preds, pred)
            uncs = append(uncs, unc)
    
    
    output=array([preds.transpose(),uncs.transpose()])
    
    vprint(["Opening output file at",args.o[0]], 1)
    try:
        outfile=open(args.o[0],"w")
    except:
        print "ERROR: Failed to open output CSV file at ",args.o[0]
        sys.exit(0)
    savetxt(outfile,output,delimiter=",")
    
    
    if verbose_level > 0 :
        print "Program time=",time.time()-start,"seconds"
    sys.exit(0)

